from is_core.rest.resource import RESTModelResource


def modelrest_factory(model, resource_class=RESTModelResource, register=True, resource_kwargs=None):
    resource_kwargs = resource_kwargs or {}
    resource_kwargs.update({
        'model': model,
        'register': register,
    })
    class_name = model.__name__ + str('Resource')
    return type(resource_class)(class_name, (resource_class,), resource_kwargs)



from django.contrib.auth.models import User
from django.test import RequestFactory

from germanium.test_cases.default import GermaniumTestCase
from germanium.tools import assert_equal, assert_raises, assert_is_none

from is_core.utils import (
    get_field_label_from_path, get_field_from_model_or_none, get_field_widget_from_path,
    get_readonly_field_value_from_path
)
from is_core.forms.utils import ReadonlyValue
from is_core.utils.field_api import (
    GetFieldDescriptorException, get_field_value_from_path, GetFieldDescriptorValueError,
    get_field_descriptors_from_path
)
from is_core.forms.widgets import (
    ReadonlyWidget, ManyToManyReadonlyWidget, ModelObjectReadonlyWidget, ModelMultipleReadonlyWidget,
    ModelChoiceReadonlyWidget
)

from issue_tracker.cores.views import UserDetailView
from issue_tracker.models import Issue

from .factories import IssueFactory, UserFactory


__all__ =(
    'UtilsTestCase',
)


class UtilsTestCase(GermaniumTestCase):

    def set_up(self):
        self.factory = RequestFactory()

    def test_get_field_label_from_path_should_return_right_descriptors(self):
        issue_name_descriptors = get_field_descriptors_from_path(Issue, 'name')
        assert_equal(len(issue_name_descriptors), 1)
        assert_equal(issue_name_descriptors[0].field_name, 'name')

        issue_solver_watching_issues_count_descriptors = get_field_descriptors_from_path(
            Issue, 'solver__watching_issues_count'
        )
        assert_equal(len(issue_solver_watching_issues_count_descriptors), 2)
        assert_equal(issue_solver_watching_issues_count_descriptors[0].field_name, 'solver')
        assert_equal(issue_solver_watching_issues_count_descriptors[1].field_name, 'watching_issues_count')

    def test_get_field_label_from_path_should_return_right_field_label(self):
        assert_equal(get_field_label_from_path(Issue, 'name'), 'Name')
        assert_equal(get_field_label_from_path(Issue, 'name', field_labels={'name': 'another name'}), 'another name')
        assert_equal(get_field_label_from_path(User, 'created_issues_count'), 'created issues count')
        assert_equal(get_field_label_from_path(Issue, '_obj_name'), 'object name')
        assert_equal(get_field_label_from_path(Issue, 'solver___obj_name'), 'Solver')
        assert_equal(
            get_field_label_from_path(Issue, 'solver___obj_name', field_labels={'solver': None}), 'object name'
        )
        assert_equal(
            get_field_label_from_path(Issue, 'solver___obj_name', field_labels={'solver__': None}), 'object name'
        )
        assert_equal(
            get_field_label_from_path(Issue, 'solver', field_labels={'solver': None}), ''
        )
        assert_equal(
            get_field_label_from_path(Issue, 'solver__watching_issues_count'), 'Solver - watching count'
        )
        assert_equal(
            get_field_label_from_path(User, 'leading_issue_name', view=UserDetailView()), 'leading issue name'
        )

    def test_get_field_label_from_path_should_raise_exception_for_invalid_field_name(self):
        with assert_raises(GetFieldDescriptorException):
            get_field_label_from_path(Issue, 'invalid')

        with assert_raises(GetFieldDescriptorException):
            get_field_label_from_path(Issue, 'solver__invalid')

    def test_get_field_from_model_or_none_should_return_model_field(self):
        assert_equal(get_field_from_model_or_none(Issue, 'name'), Issue._meta.get_field('name'))
        assert_equal(get_field_from_model_or_none(Issue, 'solver'), Issue._meta.get_field('solver'))

    def test_get_field_from_model_or_none_should_return_none_for_missing_field(self):
        assert_is_none(get_field_from_model_or_none(Issue, 'invalid'))

    def test_get_field_widget_from_path_should_return_right_wiget(self):
        assert_equal(get_field_widget_from_path(Issue, 'name'), ReadonlyWidget)
        assert_equal(get_field_widget_from_path(User, 'created_issues'), ManyToManyReadonlyWidget)
        assert_equal(get_field_widget_from_path(Issue, 'watched_by'), ModelMultipleReadonlyWidget)
        assert_equal(get_field_widget_from_path(Issue, 'solver'), ModelChoiceReadonlyWidget)
        assert_equal(get_field_widget_from_path(User, 'solving_issue'), ModelObjectReadonlyWidget)
        assert_equal(get_field_widget_from_path(User, 'solving_issue__name'), ReadonlyWidget)
        assert_equal(get_field_widget_from_path(Issue, 'related_object'), ReadonlyWidget)

    def test_get_field_widget_from_path_should_raise_exception_for_invalid_field_name(self):
        with assert_raises(GetFieldDescriptorException):
            get_field_widget_from_path(Issue, 'invalid')

    def test_get_field_value_from_path_should_return_right_value_from_model_instance(self):
        solver = UserFactory()
        leader = UserFactory()
        issue = IssueFactory(solver=solver, leader=leader, created_by=leader)
        issue.watched_by.add(solver, leader)
        assert_equal(get_field_value_from_path(issue, 'name'), issue.name)
        assert_equal(get_field_value_from_path(issue, 'is_issue'), True)
        assert_equal(
            get_field_value_from_path(issue, 'watched_by_string'), ', '.join([str(u) for u in (solver, leader)])
        )
        assert_equal(get_field_value_from_path(issue, 'created_by__first_name'), leader.first_name)
        assert_equal(
            get_field_value_from_path(issue, 'watched_by_method__first_name'), [solver.first_name, leader.first_name]
        )

    def test_get_field_value_from_path_should_return_right_value_from_resource_method(self):
        solver = UserFactory()
        leader = UserFactory()
        issue = IssueFactory(solver=solver, leader=leader, created_by=leader)
        issue.watched_by.add(solver, leader)
        request = self.factory.get('')
        with assert_raises(GetFieldDescriptorValueError):
            get_field_value_from_path(issue, 'solver__watching_issues_count')
        assert_equal(get_field_value_from_path(issue, 'solver__watching_issues_count', request=request), 1)

    def test_get_field_value_from_path_should_return_right_value_from_core_method(self):
        solver = UserFactory()
        leader = UserFactory()
        issue = IssueFactory(solver=solver, leader=leader, created_by=leader)
        issue.watched_by.add(solver, leader)
        request = self.factory.get('')
        with assert_raises(GetFieldDescriptorValueError):
            get_field_value_from_path(issue, 'leader__created_issues_count')
        assert_equal(get_field_value_from_path(issue, 'leader__created_issues_count', request=request), 1)

    def test_get_field_value_from_path_should_return_right_value_from_view_method(self):
        assert_equal(
            get_field_value_from_path(UserFactory(), 'leading_issue_name', view=UserDetailView()), 'No leading issue'
        )

    def test_get_field_value_from_path_should_raise_exception_invalid_field_name(self):
        with assert_raises(GetFieldDescriptorException):
            get_field_label_from_path(Issue, 'invalid')

    def test_get_readonly_field_value_from_path_should_return_right_value_from_model_instance(self):
        solver = UserFactory()
        leader = UserFactory()
        issue = IssueFactory(solver=solver, leader=leader, created_by=leader, related_object=solver)
        issue.watched_by.add(solver, leader)
        assert_equal(get_readonly_field_value_from_path(issue, 'name'), issue.name, None)
        assert_equal(get_readonly_field_value_from_path(issue, 'is_issue'), True)
        assert_equal(
            get_readonly_field_value_from_path(issue, 'watched_by_string'),
            ', '.join([str(u) for u in (solver, leader)])
        )
        assert_equal(get_readonly_field_value_from_path(issue, 'created_by__first_name'), leader.first_name)
        assert_equal(
            list(
                get_readonly_field_value_from_path(
                    issue, 'watched_by_method__first_name'
                )
            ),
            [solver.first_name, leader.first_name]
        )
        assert_equal(get_readonly_field_value_from_path(issue, 'related_object'), solver, None)



from is_core.rest.resource import RESTResource, RESTModelCoreMixin, RESTModelResource


class NumberOfUserIssuesResource(RESTModelCoreMixin, RESTResource):

    def _get_pk(self):
        return self.kwargs.get(self.pk_name)

    def get(self):
        user = self._get_obj_or_404()
        return {
            'created': user.created_issues.count(),
            'watching': user.watching_issues.count(),
        }


class UserModelResource(RESTModelResource):

    def watching_issues_count(self, obj):
        return obj.watching_issues.count()
    watching_issues_count.short_description = 'watching count'



from django.db.models import Model
from django.utils.functional import lazy

from django.apps import apps
from django.db.models.utils import make_model_tuple


def short_description(description):
    """
    Sets 'short_description' attribute (this attribute is in exports to generate header name).
    """
    def decorator(func):
        if isinstance(func, property):
            func = func.fget
        func.short_description = description
        return func
    return decorator


def relation(to):
    """
    Add relation to the method. Now method can be used in the core field path as a relation.
    :param to: model class or string path to the model class
    """
    def decorator(func):
        if isinstance(func, property):
            func = func.fget

        def set_related_model(model):
            func.related_model = model

        if isinstance(to, str):
            apps.lazy_model_operation(set_related_model, make_model_tuple(to))
        else:
            func.related_model = to
        return func
    return decorator



import json

from django.http.response import HttpResponse


class JsonHttpResponse(HttpResponse):
    status_code = 200

    def __init__(self, content=None, *args, **kwargs):
        super(HttpResponse, self).__init__(content_type='application/json', *args, **kwargs)
        self.content = json.dumps(content)


class JsonCreatedHttpResponse(JsonHttpResponse):
    status_code = 201



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-08-18 18:10
from __future__ import unicode_literals

import import_string

import chamber.models.fields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import is_core.contrib.background_export.models
from is_core.config import settings as is_core_settings


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='ExportedFile',
            fields=[
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('slug', models.SlugField(max_length=32, primary_key=True, serialize=False, verbose_name='slug')),
                ('file', chamber.models.fields.FileField(
                    blank=True, null=True, upload_to=is_core.contrib.background_export.models.generate_filename,
                    verbose_name='file', storage=import_string(is_core_settings.BACKGROUND_EXPORT_STORAGE_CLASS)()
                )),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='contenttypes.ContentType')),
                ('created_by',
                 models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='created_exported_files',
                                   to=settings.AUTH_USER_MODEL, verbose_name='created by')),
                ('downloaded_by', models.ManyToManyField(blank=True, related_name='downloaded_exported_files',
                                                         to=settings.AUTH_USER_MODEL, verbose_name='downloaded by')),
            ],
            options={
                'verbose_name': 'exported file',
                'verbose_name_plural': 'exported files',
                'ordering': ('-created_at',),
            },
        ),
    ]



import import_string

from django.conf import settings as django_settings
from django.contrib.contenttypes.models import ContentType
from django.core.files.base import ContentFile
from django.http.request import HttpRequest
from django.utils import translation
from django.db.transaction import atomic

from pyston.converters import get_converter_from_request
from pyston.serializer import get_serializer, get_resource_or_none
from pyston.utils import RFS
from pyston.utils.helpers import QuerysetIteratorHelper

from is_core.config import settings

from django_celery_extensions.task import string_to_obj

from security.task import LoggedTask

from celery import shared_task

from .models import ExportedFile

from .signals import export_success


def get_rest_request(user, rest_context):
    request = HttpRequest()
    request._rest_context = rest_context
    request.user = user
    request.kwargs = {}
    return request


class FileBackgroundExportGenerator:

    def __init__(self, model):
        self.model = model

    def generate(self, exported_file, request, queryset, requested_fieldset, serialization_format):
        converter = get_converter_from_request(request)
        converted_dict = get_serializer(queryset, request=request).serialize(
            QuerysetIteratorHelper(queryset), serialization_format, requested_fieldset=requested_fieldset,
            lazy=True, allow_tags=converter.allow_tags
        )
        django_file = exported_file.file
        try:
            django_file.open('wb')
            converter.encode_to_stream(
                django_file.file, converted_dict, resource=get_resource_or_none(request, queryset.model),
                request=request, requested_fields=requested_fieldset
            )
        finally:
            django_file.close()


class BackgroundSerializationTask(LoggedTask):

    abstract = True

    def get_exported_file(self, pk):
        return ExportedFile.objects.get(pk=pk)

    def on_task_success(self, task_id, args, kwargs, retval):
        super().on_task_success(task_id, args, kwargs, retval)
        exported_file = self.get_exported_file(args[0])
        export_success.send(sender=self.__class__, exported_file=exported_file)


@shared_task(base=BackgroundSerializationTask,
             name='background_export.serializer.serialization',
             queue=settings.BACKGROUND_EXPORT_TASK_QUEUE,
             time_limit=settings.BACKGROUND_EXPORT_TASK_TIME_LIMIT,
             soft_time_limit=settings.BACKGROUND_EXPORT_TASK_SOFT_TIME_LIMIT,
             bind=True)
@atomic
def background_serialization(self, exported_file_pk, rest_context, language, requested_fieldset, serialization_format,
                             filename, query):
    # Must be here, because handlers is not registered
    import_string(django_settings.ROOT_URLCONF)

    prev_language = translation.get_language()
    translation.activate(language)
    try:
        exported_file = self.get_exported_file(exported_file_pk)
        exported_file.file.save(filename, ContentFile(''))
        request = get_rest_request(exported_file.created_by, rest_context)
        if settings.BACKGROUND_EXPORT_TASK_UPDATE_REQUEST_FUNCTION:
            request = import_string(settings.BACKGROUND_EXPORT_TASK_UPDATE_REQUEST_FUNCTION)(request)
        query = string_to_obj(query)
        queryset = query.model.objects.all()
        queryset.query = query
        FileBackgroundExportGenerator(query.model).generate(
            exported_file, request, queryset, RFS.create_from_string(requested_fieldset), serialization_format
        )
    finally:
        translation.activate(prev_language)



from collections import OrderedDict

from django.db.models import Model
from django.conf.urls import url, include
from django.template.defaultfilters import lower
from django.core.exceptions import ImproperlyConfigured

import import_string

from is_core.config import settings

from .loading import get_cores
from .patterns import RESTPattern
from .rest.resource import EntryPointResource


sites = {}
registered_model_cores = {}
registered_cores = []


def get_site_by_name(name):
    """
    Return site according to name (default is IS)
    """
    return sites.get(name)


def get_model_core(model):
    """
    Return core view of given model or None
    """
    return registered_model_cores.get(model)


class ISSite:

    def __init__(self, name='IS'):
        self.name = name
        sites[name] = self
        self._registry = self._init_items()

    def _init_items(self):
        out = OrderedDict()

        for core in get_cores():
            generic_core = self.register(core(self.name, []))
            if generic_core.menu_group in out:
                raise ImproperlyConfigured('Duplicate cores with group: "%s"' % generic_core.menu_group)
            out[generic_core.menu_group] = generic_core
        return out

    def register(self, generic_core):
        if getattr(generic_core, 'register_model', False):
            registered_model_cores[generic_core.model] = generic_core
        registered_cores.append(generic_core)
        return generic_core

    @property
    def urls(self):
        return self.get_urls(), self.name

    def _set_items_urls(self, items, urlpatterns):
        for item in items:
            urls = item.get_urls()
            if urls:
                urlpatterns += [url(r'^', include(urls))]

    def get_urls(self):
        urlpatterns = []
        if settings.AUTH_LOGIN_VIEW:
            urlpatterns.append(
                url(r'^{}$'.format(settings.LOGIN_URL[1:]), import_string(settings.AUTH_LOGIN_VIEW).as_view(
                    form_class=import_string(settings.AUTH_FORM_CLASS)), name='login')
            )

        if settings.AUTH_LOGOUT_VIEW:
            urlpatterns.append(
                url(r'^{}$'.format(settings.LOGOUT_URL[1:]),
                    import_string(settings.AUTH_LOGOUT_VIEW).as_view(), name='logout')
            )

        if settings.AUTH_RESOURCE_CLASS:
            auth_resource_class = import_string(settings.AUTH_RESOURCE_CLASS)
            auth_resource_class.form_class = import_string(settings.AUTH_FORM_CLASS)
            pattern = RESTPattern('api-login', self.name, settings.LOGIN_API_URL[1:],
                                  auth_resource_class)
            urlpatterns.append(pattern.get_url())

        if settings.AUTH_LOGIN_CODE_VERIFICATION_VIEW:
            urlpatterns.append(
                url(r'^{}$'.format(settings.CODE_VERIFICATION_URL[1:]),
                    import_string(settings.AUTH_LOGIN_CODE_VERIFICATION_VIEW).as_view(), name='code-verification-login')
            )

        pattern = RESTPattern('api', self.name, r'api/', EntryPointResource)
        urlpatterns.append(pattern.get_url())

        self._set_items_urls(self._registry.values(), urlpatterns)
        return urlpatterns


site = ISSite()


def get_cores():
    return site._registry.values()


def get_core(name):
    return site._registry[name]



from django.db.models.query import QuerySet

from pyston.paginator import OffsetBasedPaginator as OriginOffsetBasedPaginator
from pyston.paginator import OffsetBasedPaginatorWithoutTotal as OriginOffsetBasedPaginatorWithoutTotal
from pyston.paginator import CursorBasedPaginator as OriginCursorBasedPaginator

from is_core.config import settings


class OffsetBasedPaginatorMixin:

    type = 'offset-based-paginator'


class OffsetBasedPaginator(OffsetBasedPaginatorMixin, OriginOffsetBasedPaginator):

    type = 'offset-based-paginator'

    def _get_total(self, qs, request):
        if request._rest_context.get('request_count', False):
            return super()._get_total(qs, request)
        else:
            return None


class CursorBasedPaginator(OriginCursorBasedPaginator):

    type = 'cursor-based-paginator'



import os

from setuptools import setup, find_packages

from is_core.version import get_version


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='django-is-core',
    version=get_version(),
    description="Information systems core.",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    keywords='django, admin, information systems, REST',
    author='Lubos Matl',
    author_email='matllubos@gmail.com',
    url='https://github.com/matllubos/django-is-core',
    license='BSD',
    package_dir={'is_core': 'is_core'},
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ],
    install_requires=[
        'django>=2.2',
        'django-pyston>=2.14.9',
        'django-block-snippets>=2.0.1',
        'django-chamber>=0.6.11',
        'python-dateutil>=2.8.1',
        'pytz',
        'Unidecode',
        'import_string>=0.1.0',
    ],
    zip_safe=False
)



from is_core.generic_views.form_views import DetailModelFormView


class UserDetailView(DetailModelFormView):

    def leading_issue_name(self, obj):
        return obj.leading_issue.name if hasattr(obj, 'leading_issue') else 'No leading issue'
    leading_issue_name.short_description = 'leading issue name'



from django.utils.html import format_html


def add_class_name(attrs, class_name):
    class_names = attrs.get('class')
    if class_names:
        class_names = [class_names]
    else:
        class_names = []
    class_names.append(class_name)
    attrs['class'] = ' '.join(class_names)
    return attrs


class ReadonlyValue:

    def __init__(self, value, humanized_value):
        self._value = value
        self._humanized_value = humanized_value

    def render(self):
        if self._humanized_value is None:
            return format_html(
                '<p>{}</p>',
                self._value,
            )
        else:
            return format_html(
                '<p><span title="{}">{}</span></p>',
                self._value,
                self.value
            )

    @property
    def value(self):
        return self._humanized_value if self._humanized_value is not None else self._value